/************************************************************************/
/* Author: Cheng Ren(γ), Beihang Univeristy
   Date  : 2010/12/13
   Description: For some engineers in computer graphics area, analyze a group of position vectors sometimes is necessary.
                This source file is used to analyze the principle components of such specific type of data: Vector3(x,y,z).
                Use parameter DesiredDimension to specify which dimension you want to compress to, for example 1 dimension,
				2 dimensions. If you specify 3 dimensions, then the result data will the same as your input data.
   Reference : Lindsay I Smith, A tutorial on principle component analysis.
   */
/************************************************************************/

#include "Vector3.h"
#include "Matrix3.h"
#include <vector>
#include <fstream>
#include <iostream>
using namespace std;
using namespace PCA;

PCA::Vector3 CalcMean( const std::vector<PCA::Vector3>& Points )
{
	PCA::Vector3 mean(0,0,0);
	for( std::vector<PCA::Vector3>::const_iterator _iter=Points.begin(); _iter!=Points.end(); _iter++ )
	{
		mean += *_iter;
	}
	mean /= Points.size();
	return mean;
}

float CalcMean( const std::vector<float>& Points )
{
	float mean = 0;
	for( std::vector<float>::const_iterator _iter=Points.begin(); _iter!=Points.end(); _iter++ )
	{
		mean += *_iter;
	}
	mean /= Points.size();
	return mean;
}

float CalcCovavriance( const std::vector<float>& PointsDim1, const std::vector<float>& PointsDim2 )
{
	assert( PointsDim1.size() == PointsDim2.size() );
	unsigned int DataSize = PointsDim1.size();
	if( DataSize ==0 )
		return 0;

	float mean1 = CalcMean( PointsDim1 );
	float mean2 = CalcMean( PointsDim2 );

	float Covariance = 0.f;
	for( unsigned int i=0; i<DataSize; i++ )
	{
		Covariance += (PointsDim1[i] - mean1) * (PointsDim2[i] - mean2);
	}
	Covariance /= DataSize-1;
	return Covariance;
}

void CalcCovavrianceMatrix( const std::vector<PCA::Vector3>& Points, PCA::Matrix3& outMatrix )
{
	std::vector<float> PointsDimX;
	std::vector<float> PointsDimY;
	std::vector<float> PointsDimZ;
	PointsDimX.reserve( Points.size() );
	PointsDimY.reserve( Points.size() );
	PointsDimZ.reserve( Points.size() );
	for(std::vector<PCA::Vector3>::const_iterator _iter=Points.begin(); _iter!=Points.end(); _iter++ )
	{
		PointsDimX.push_back( _iter->x );
		PointsDimY.push_back( _iter->y );
		PointsDimZ.push_back( _iter->z );
	}

	outMatrix[0][0] = CalcCovavriance( PointsDimX, PointsDimX );
	outMatrix[1][1] = CalcCovavriance( PointsDimY, PointsDimY );
	outMatrix[2][2] = CalcCovavriance( PointsDimZ, PointsDimZ );

	outMatrix[0][1] = outMatrix[1][0] = CalcCovavriance( PointsDimX, PointsDimY );
	outMatrix[0][2] = outMatrix[2][0] = CalcCovavriance( PointsDimX, PointsDimZ );
	outMatrix[1][2] = outMatrix[2][1] = CalcCovavriance( PointsDimY, PointsDimZ );
}

void CalcEigens(PCA::Matrix3 SymmetricMatrix, PCA::Real outEigenvalue[3], PCA::Vector3 outEigenvector[3])
{
	PCA::Real Eigenvalue[3];
	PCA::Vector3 Eigenvector[3];

	SymmetricMatrix.EigenSolveSymmetric( Eigenvalue, Eigenvector );

	// sort the Eigens according to their importances( from most important to least important ).
	int MaxEigenIndex = ( Eigenvalue[0] >= Eigenvalue[1] ) ? 0 : 1;
	MaxEigenIndex = ( Eigenvalue[MaxEigenIndex] > Eigenvalue[2] ) ? MaxEigenIndex : 2;

	int MinEigenIndex = ( Eigenvalue[0] < Eigenvalue[1] ) ? 0 : 1;
	MinEigenIndex = ( Eigenvalue[MinEigenIndex] < Eigenvalue[2] ) ? MinEigenIndex : 2;

	int temp = MaxEigenIndex + MinEigenIndex;
	int TrickSign = 2 - temp;
	int SecondEigenIndex = 0;
	if( TrickSign==0 )
		SecondEigenIndex = 1;
	else if ( TrickSign == 1 )
		SecondEigenIndex = 2;

	outEigenvalue[0] = Eigenvalue[MaxEigenIndex];
	outEigenvalue[1] = Eigenvalue[SecondEigenIndex];
	outEigenvalue[2] = Eigenvalue[MinEigenIndex];

	outEigenvector[0] = Eigenvector[MaxEigenIndex];
	outEigenvector[1] = Eigenvector[SecondEigenIndex];
	outEigenvector[2] = Eigenvector[MinEigenIndex];
}

/// Param @DesiredDimension : to which dimension you want to compress. The range is [1,3].
void PCAAnalyze( const std::vector<PCA::Vector3>& OriginalData, unsigned int DesiredDimension, std::vector<PCA::Vector3>& ResultData, PCA::Real outEigenvalue[3], PCA::Vector3 outEigenvector[3] )
{
	if( DesiredDimension<0 || DesiredDimension>3 )
		return;

	// Step1 : Get the data, which is OriginalData

	// Step2 : Subtract the mean
	PCA::Vector3 mean = CalcMean( OriginalData );
	std::vector<PCA::Vector3> SubtractedPoints;
	SubtractedPoints.reserve( OriginalData.size() );
	for( std::vector<PCA::Vector3>::const_iterator _iter=OriginalData.begin(); _iter!=OriginalData.end(); _iter++ )
	{
		SubtractedPoints.push_back( *_iter - mean );
	}

	// Step3 : Caculate the covariance matrix
	PCA::Matrix3 CovMatrix;
	CalcCovavrianceMatrix( SubtractedPoints, CovMatrix );

	// Step4 : Caculate the Eigens
	CalcEigens( CovMatrix , outEigenvalue, outEigenvector );

	// Step5 ; Deriving the new data set
	int DataSize = OriginalData.size();
	std::vector<PCA::Vector3> DerivedData;
	DerivedData.reserve( OriginalData.size() );
	for( std::vector<PCA::Vector3>::const_iterator _iter=OriginalData.begin(); _iter!=OriginalData.end(); _iter++ )
	{
		DerivedData.push_back( PCA::Vector3::ZERO );
	}
	for( unsigned int i=0; i<DesiredDimension; i++ )
	{
		for( int iDataIndex=0; iDataIndex<DataSize; iDataIndex++ )
		{
			DerivedData[iDataIndex][i] = outEigenvector[i].x*SubtractedPoints[iDataIndex].x + outEigenvector[i].y*SubtractedPoints[iDataIndex].y + outEigenvector[i].z*SubtractedPoints[iDataIndex].z;
		}
	}

	// Get the final Data based on the pca derived data
	ResultData.clear();
	ResultData.reserve( DerivedData.size() );
	for( int iDataIndex=0; iDataIndex<DataSize; iDataIndex++ )
	{
		ResultData.push_back( mean );
	}
	for( unsigned int i=0; i<DesiredDimension; i++ )
	{

		for( int iDataIndex=0; iDataIndex<DataSize; iDataIndex++ )
		{
			ResultData[iDataIndex] += outEigenvector[i] * DerivedData[iDataIndex][i];
		}
	}
}

int main(int argc, char* argv[])
{
	// Read the input file to get sampled data
	std::vector<PCA::Vector3> SampleData;
	std::ifstream SampleTxt("SampleData.txt");
	if( ! SampleTxt.is_open() )
	{
		std::cout << "Error! : Couldn't find SampleData.txt" << std::endl;
		return 0;
	}
	PCA::Vector3 Samples;
	char buffer[1024];
	while( !SampleTxt.eof() )
	{
		SampleTxt.getline(buffer, 1024);
		if( buffer[0] == '\0' )
			break;
		sscanf( buffer, "%f %f %f", &Samples.x, &Samples.y, &Samples.z );
		SampleData.push_back( Samples );
	}
	SampleTxt.close();

	// Output sampled data to a 3DS Max script file, so that we can visualize the data in 3DS Max.
	int DataSize = SampleData.size();
	std::ofstream OriginalDataTxt( "Results\\DisplaySamples.ms" );
	for( int i=0; i<DataSize; i++ )
	{
		OriginalDataTxt << "PointHelperObj box:true size:0.01 name:\"helper_" << i << "\"" << " pos:[" << SampleData[i].x << "," << SampleData[i].y << "," << SampleData[i].z << "]" << std::endl;
	}
	OriginalDataTxt.close();


	//------------------Core function : Use PCA to anaylze the given data
	std::vector<PCA::Vector3> ResultData;
	PCA::Real Eigenvalue[3];
	PCA::Vector3 Eigenvector[3];
	PCAAnalyze( SampleData, 2, ResultData, Eigenvalue, Eigenvector);
	//------------------Core function End

	std::ofstream AnalyzedDataTxt( "Results\\PCAResult.txt" );
	AnalyzedDataTxt.precision(6);
	for( int i=0; i<DataSize; i++ )
	{
		AnalyzedDataTxt << ResultData[i].x << "\t" << ResultData[i].y << "\t" << ResultData[i].z << std::endl;
	}
	AnalyzedDataTxt.close();

	// Output the result points to a 3DS Max script file, so that we can visualize the data in 3DS Max.
	std::ofstream DispAnalyzedDataTxt( "Results\\DisplayResult.ms" );
	for( int i=0; i<DataSize; i++ )
	{
		DispAnalyzedDataTxt << "PointHelperObj box:true size:0.01 name:\"helper_" << i << "\"" << " pos:[" << ResultData[i].x << "," << ResultData[i].y << "," << ResultData[i].z << "]" << std::endl;
	}
	// Output the result axises
	PCA::Vector3 mean = CalcMean( ResultData );
	DispAnalyzedDataTxt << "ss = SplineShape() " << std::endl;
	DispAnalyzedDataTxt << "addNewSpline ss" << std::endl;
	DispAnalyzedDataTxt << "addKnot ss 1 #corner #line [" << mean.x+10*Eigenvector[0].x << "," << mean.y+10*Eigenvector[0].y << "," << mean.z+10*Eigenvector[0].z << "]" << std::endl;
	DispAnalyzedDataTxt << "addKnot ss 1 #corner #line [" << mean.x << "," << mean.y << "," << mean.z << "]" << std::endl;
	DispAnalyzedDataTxt << "addKnot  ss 1 #corner #line [" << mean.x+10*Eigenvector[1].x << "," << mean.y+5*Eigenvector[1].y << "," << mean.z+5*Eigenvector[1].z << "]" <<std::endl;
	DispAnalyzedDataTxt << "updateshape ss" << std::endl;
	DispAnalyzedDataTxt.close();

	std::cout << "The result data has been saved to text file PCAResult.txt." << std::endl << std::endl;
	std::cout << "The original data and the result data have also been saved to script files DisplaySamples.ms and DisplayResult.ms separately." << std::endl << std::endl;
	std::cout << "It is recommended that you use 3DS Max to run the .ms script files, so that you visually observe the data." << std::endl << std::endl;
}